package ch.usi.inf.sa4.sanmarinoes.smarthut.service;

import ch.usi.inf.sa4.sanmarinoes.smarthut.models.Thermostat;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.ThermostatRepository;
import ch.usi.inf.sa4.sanmarinoes.smarthut.socket.SensorSocketEndpoint;
import ch.usi.inf.sa4.sanmarinoes.smarthut.utils.Utils;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.List;
import java.util.Optional;
import java.util.Random;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class ThermostatService {

    private final SensorSocketEndpoint endpoint;

    private final DeviceService deviceService;

    private final ThermostatPopulationService thermostatPopulationService;

    private final ThermostatRepository thermostatRepository;

    private Random ran = new Random();

    @Autowired
    public ThermostatService(
            SensorSocketEndpoint endpoint,
            DeviceService deviceService,
            ThermostatPopulationService thermostatPopulationService,
            ThermostatRepository thermostatRepository) {
        this.endpoint = endpoint;
        this.deviceService = deviceService;
        this.thermostatPopulationService = thermostatPopulationService;
        this.thermostatRepository = thermostatRepository;
    }

    private void randomJitter(Thermostat thermostat) {

        BigDecimal x =
                thermostat
                        .getTypical()
                        .subtract(
                                thermostat
                                        .getErr()
                                        .divide(BigDecimal.valueOf(2), RoundingMode.CEILING))
                        .add(
                                BigDecimal.valueOf(
                                        ran.nextDouble() * thermostat.getErr().doubleValue() * 2));

        updateValueForThermostat(thermostat, x);
    }

    private void updateValueForThermostat(Thermostat thermostat, BigDecimal value) {
        thermostat.setInternalSensorTemperature(value);
        deviceService.saveAsOwner(
                thermostat, thermostatRepository.findUser(thermostat.getId()).getUsername());
    }

    public void fakeUpdateAll() {
        thermostatRepository.findAll().forEach(this::randomJitter);
        updateStates();
    }

    public List<Thermostat> findAll(String username) {
        Iterable<Thermostat> all = thermostatRepository.findAllByUsername(username);
        all.forEach(thermostatPopulationService::populateMeasuredTemperature);
        return Utils.toList(all);
    }

    public void computeState(Thermostat t) {
        thermostatPopulationService.populateMeasuredTemperature(t);
        t.computeState();
    }

    private void updateState(Thermostat t) {
        this.computeState(t);

        deviceService.saveAsOwner(t, thermostatRepository.findUser(t.getId()).getUsername());
        endpoint.queueDeviceUpdate(t, thermostatRepository.findUser(t.getId()), false, null, false);
    }

    public void updateStates() {
        Iterable<Thermostat> ts = thermostatRepository.findAll();
        ts.forEach(this::updateState);
    }

    public Optional<Thermostat> findById(Long thermostat, String username) {
        Optional<Thermostat> t = thermostatRepository.findByIdAndUsername(thermostat, username);

        if (t.isPresent()) {
            Thermostat u = t.get();
            thermostatPopulationService.populateMeasuredTemperature(u);
            t = Optional.of(u);
        }
        return t;
    }
}
