package ch.usi.inf.sa4.sanmarinoes.smarthut.controller;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.when;

import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.SwitchableSaveRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.DuplicateStateException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.NotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.*;
import ch.usi.inf.sa4.sanmarinoes.smarthut.service.DeviceService;
import java.security.Principal;
import java.util.List;
import java.util.Optional;
import lombok.SneakyThrows;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.mock.web.MockHttpServletRequest;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

@ExtendWith(MockitoExtension.class)
@WithMockUser(username = "user")
public class RegularLightControllerTests {

    @InjectMocks private RegularLightController regularLightController;

    @Mock private RegularLightRepository regularLightRepository;

    @Mock private DeviceService deviceService;

    @Mock private Principal mockPrincipal;

    @Mock private UserRepository userRepository;

    @Mock private SceneRepository sceneRepository;

    @Mock private StateRepository<State> stateRepository;

    private void checkRegularLightAgainstRequest(
            final RegularLight toCheck, final SwitchableSaveRequest request) {
        assertThat(toCheck).isNotNull();
        assertThat(toCheck.isOn()).isEqualTo(request.isOn());
        assertThat(toCheck.getName()).isEqualTo(request.getName());
        assertThat(toCheck.getRoomId()).isEqualTo(request.getRoomId());
    }

    @Test
    public void testGetAll() {
        when(regularLightRepository.findAll()).thenReturn(List.of());
        assertThat(regularLightController.findAll()).isEmpty();
    }

    @Test
    public void testGet() throws NotFoundException {
        RegularLight r = new RegularLight();
        when(regularLightRepository.findById(1L)).thenReturn(Optional.of(r));
        when(regularLightRepository.findById(2L)).thenReturn(Optional.empty());

        assertThat(regularLightController.findById(1L)).isSameAs(r);
        assertThatThrownBy(() -> regularLightController.findById(2L))
                .isInstanceOf(NotFoundException.class);
    }

    @Test
    @DisplayName("when creating should return the same object")
    @SneakyThrows(NotFoundException.class)
    public void testCreate() {
        when(mockPrincipal.getName()).thenReturn("user");
        doNothing().when(deviceService).throwIfRoomNotOwned(anyLong(), eq("user"));
        when(deviceService.saveAsOwner(any(RegularLight.class), eq("user")))
                .thenAnswer(i -> i.getArguments()[0]);

        MockHttpServletRequest request = new MockHttpServletRequest();
        RequestContextHolder.setRequestAttributes(new ServletRequestAttributes(request));

        final SwitchableSaveRequest toSend = new SwitchableSaveRequest();
        toSend.setName("rl");
        toSend.setRoomId(20L);
        toSend.setOn(true);

        final RegularLight regularLight = regularLightController.create(toSend, mockPrincipal);

        checkRegularLightAgainstRequest(regularLight, toSend);
    }

    @Test
    @DisplayName("when updating should return the same object")
    @SneakyThrows(NotFoundException.class)
    public void testUpdate() {

        when(mockPrincipal.getName()).thenReturn("user");
        final SwitchableSaveRequest toSend = new SwitchableSaveRequest();
        toSend.setName("rl");
        toSend.setRoomId(20L);
        toSend.setOn(true);

        final RegularLight toUpdate = new RegularLight();
        toUpdate.setName("OOO");
        toUpdate.setRoomId(40L);
        toUpdate.setOn(false);

        when(regularLightRepository.findByIdAndUsername(50L, "user"))
                .thenReturn(Optional.of(toUpdate));
        when(regularLightRepository.findByIdAndUsername(60L, "user")).thenReturn(Optional.empty());

        when(deviceService.saveAsOwner(toUpdate, "user")).thenReturn(toUpdate);

        toSend.setId(60L);
        assertThatThrownBy(() -> regularLightController.update(toSend, mockPrincipal, null))
                .isInstanceOf(NotFoundException.class);
        toSend.setId(50L);
        assertThat(regularLightController.update(toSend, mockPrincipal, null)).isEqualTo(toUpdate);

        toUpdate.setName("OOO");
        toUpdate.setRoomId(40L);
        toSend.setId(50L);
        toUpdate.setOn(false);

        when(regularLightRepository.findByIdAndUserId(anyLong(), anyLong()))
                .thenReturn(java.util.Optional.of(toUpdate));

        when(deviceService.saveAsGuest(any(RegularLight.class), eq("user"), anyLong()))
                .thenAnswer(i -> i.getArguments()[0]);

        User guest = new User();
        User host = new User();
        host.getGuests().add(guest);
        guest.getHosts().add(host);

        when(userRepository.findById(20L)).thenReturn(Optional.of(host));
        when(userRepository.findByUsername("user")).thenReturn(guest);

        MockHttpServletRequest request = new MockHttpServletRequest();
        RequestContextHolder.setRequestAttributes(new ServletRequestAttributes(request));

        final RegularLight regularLight = regularLightController.update(toSend, mockPrincipal, 20L);

        checkRegularLightAgainstRequest(regularLight, toSend);
    }

    @Test
    @DisplayName("an existing id should succeed")
    @SneakyThrows(NotFoundException.class)
    public void testDelete() {

        when(mockPrincipal.getName()).thenReturn("user");
        doNothing().when(deviceService).deleteByIdAsOwner(eq(42L), eq("user"));

        MockHttpServletRequest request = new MockHttpServletRequest();
        RequestContextHolder.setRequestAttributes(new ServletRequestAttributes(request));

        Assertions.assertDoesNotThrow(() -> regularLightController.delete(42L, mockPrincipal));
    }

    @Test
    public void testSceneBinding() {
        when(mockPrincipal.getName()).thenReturn("user");
        RegularLight regularLight = new RegularLight();
        when(regularLightRepository.findByIdAndUsername(24L, "user"))
                .thenReturn(Optional.of(regularLight));
        Scene scene = new Scene();
        scene.setId(1L);
        SwitchableState state = new SwitchableState();
        state.setSceneId(1L);
        State s = regularLight.cloneState();
        when(sceneRepository.findById(1L)).thenReturn(Optional.of(scene));
        when(stateRepository.countByDeviceIdAndSceneId(24L, 1L)).thenReturn(0);
        Assertions.assertDoesNotThrow(
                () -> regularLightController.sceneBinding(24L, 1L, mockPrincipal));
    }

    @Test
    public void testSceneBinding2() {
        when(mockPrincipal.getName()).thenReturn("user");
        when(mockPrincipal.getName()).thenReturn("user");
        RegularLight regularLight = new RegularLight();
        when(regularLightRepository.findByIdAndUsername(24L, "user"))
                .thenReturn(Optional.of(regularLight));
        Scene scene = new Scene();
        scene.setId(1L);
        SwitchableState state = new SwitchableState();
        state.setSceneId(1L);
        State s = regularLight.cloneState();
        when(sceneRepository.findById(1L)).thenReturn(Optional.of(scene));
        when(stateRepository.countByDeviceIdAndSceneId(24L, 1L)).thenReturn(2);
        assertThatThrownBy(() -> regularLightController.sceneBinding(24L, 1L, mockPrincipal))
                .isInstanceOf(DuplicateStateException.class);
    }
}
