package ch.usi.inf.sa4.sanmarinoes.smarthut.controller;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.when;

import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.SceneSaveRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.NotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.*;
import ch.usi.inf.sa4.sanmarinoes.smarthut.service.SceneService;
import java.security.Principal;
import java.util.List;
import java.util.Optional;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.security.test.context.support.WithMockUser;

@ExtendWith(MockitoExtension.class)
@WithMockUser(username = "user")
public class SceneControllerTests {
    @InjectMocks private SceneController sceneController;

    @Mock private UserRepository userRepository;

    @Mock private SceneRepository sceneRepository;

    @Mock private Principal mockPrincipal;

    @Mock private StateRepository<State> stateStateRepository;

    @Mock private SceneService sceneService;

    private final User u;

    private final User h;

    public SceneControllerTests() {
        u = new User();
        u.setName("user");
        u.setId(1L);

        h = new User();
        h.setId(2L);
        h.setUsername("host");

        u.getHosts().add(h);
        h.getGuests().add(u);
    }

    @Test
    public void testCopy() throws NotFoundException {
        final Scene scene = new Scene();
        final Scene copyFrom = new Scene();

        when(mockPrincipal.getName()).thenReturn("user");
        when(sceneRepository.findByIdAndUsername(1L, "user")).thenReturn(Optional.of(scene));
        when(sceneRepository.findByIdAndUsername(2L, "user")).thenReturn(Optional.empty());

        when(sceneRepository.findByIdAndUsername(10L, "user")).thenReturn(Optional.of(copyFrom));
        when(sceneRepository.findByIdAndUsername(20L, "user")).thenReturn(Optional.empty());

        assertThatThrownBy(() -> sceneController.copy(2L, 10L, mockPrincipal))
                .isInstanceOf(NotFoundException.class);
        assertThatThrownBy(() -> sceneController.copy(1L, 20L, mockPrincipal))
                .isInstanceOf(NotFoundException.class);

        when(sceneService.copyStates(scene, copyFrom)).thenReturn(List.of());

        assertThat(sceneController.copy(1L, 10L, mockPrincipal)).isEmpty();
    }

    @Test
    public void testApply() throws NotFoundException {
        final Scene s = new Scene();
        when(mockPrincipal.getName()).thenReturn("user");
        when(sceneRepository.findByIdAndUsername(1L, "user")).thenReturn(Optional.of(s));
        when(sceneRepository.findByIdAndUsername(2L, "user")).thenReturn(Optional.empty());
        when(sceneService.apply(s, "user", false)).thenReturn(List.of());

        assertThatThrownBy(() -> sceneController.apply(2L, mockPrincipal, null))
                .isInstanceOf(NotFoundException.class);
        assertThat(sceneController.apply(1L, mockPrincipal, null)).isEmpty();

        when(userRepository.findByUsername("user")).thenReturn(u);
        when(userRepository.findById(2L)).thenReturn(Optional.of(h));

        when(sceneRepository.findByIdAndUserIdAndGuestAccessEnabled(1L, 2L, true))
                .thenReturn(Optional.of(s));
        when(sceneRepository.findByIdAndUserIdAndGuestAccessEnabled(2L, 2L, true))
                .thenReturn(Optional.empty());

        when(sceneService.applyAsGuest(s, "user", 2L)).thenReturn(List.of());

        assertThatThrownBy(() -> sceneController.apply(2L, mockPrincipal, 2L))
                .isInstanceOf(NotFoundException.class);
        assertThat(sceneController.apply(1L, mockPrincipal, 2L)).isEmpty();
    }

    @Test
    public void testGetAll() throws NotFoundException {
        when(mockPrincipal.getName()).thenReturn("user");
        when(sceneRepository.findByUsername("user")).thenReturn(List.of());
        assertThat(sceneController.findAll(mockPrincipal, null)).isEmpty();

        when(userRepository.findByUsername("user")).thenReturn(u);
        when(userRepository.findById(2L)).thenReturn(Optional.of(h));
        assertThat(sceneController.findAll(mockPrincipal, 2L)).isEmpty();
    }

    private void equalToRequest(Scene created, SceneSaveRequest a) {
        if (a.getName() != null) assertThat(created.getName()).isEqualTo(a.getName());
        assertThat(created.getUserId()).isEqualTo(1L);
        assertThat(created.getIcon()).isEqualTo(a.getIcon());
        assertThat(created.isGuestAccessEnabled()).isEqualTo(a.isGuestAccessEnabled());
    }

    @Test
    public void testCreate() {
        when(mockPrincipal.getName()).thenReturn("user");
        when(userRepository.findByUsername("user")).thenReturn(u);
        when(sceneRepository.save(any())).thenAnswer(i -> i.getArguments()[0]);

        SceneSaveRequest s = new SceneSaveRequest(0, "New Scene", Icon.BATH, true);
        Scene created = sceneController.create(s, mockPrincipal);
        assertThat(created.getId()).isEqualTo(0L);
        equalToRequest(created, s);
    }

    @Test
    public void testUpdate() throws NotFoundException {
        when(mockPrincipal.getName()).thenReturn("user");
        final Scene old = new Scene();
        old.setId(42L);
        old.setUserId(1L);
        old.setName("Old Name");

        when(sceneRepository.save(any())).thenAnswer(i -> i.getArguments()[0]);
        when(sceneRepository.findByIdAndUsername(42L, "user")).thenReturn(Optional.of(old));
        when(sceneRepository.findByIdAndUsername(43L, "user")).thenReturn(Optional.empty());

        SceneSaveRequest a = new SceneSaveRequest(42L, "New Scene", Icon.BATH, true);

        Scene created = sceneController.update(a.getId(), a, mockPrincipal);
        assertThat(created.getId()).isEqualTo(42L);
        equalToRequest(created, a);

        a.setName(null);
        created = sceneController.update(a.getId(), a, mockPrincipal);
        assertThat(created.getId()).isEqualTo(42L);
        equalToRequest(created, a);

        a.setId(43L);
        assertThatThrownBy(() -> sceneController.update(a.getId(), a, mockPrincipal))
                .isInstanceOf(NotFoundException.class);
    }

    @Test
    public void testDelete() {
        doNothing().when(stateStateRepository).deleteAllBySceneId(42L);
        doNothing().when(sceneRepository).deleteById(42L);
        Assertions.assertDoesNotThrow(() -> sceneController.deleteById(42L));
    }

    @Test
    public void testGetStates() {
        when(stateStateRepository.findBySceneId(1L)).thenReturn(List.of());
        assertThat(sceneController.getStates(1L)).isEmpty();
    }
}
