package ch.usi.inf.sa4.sanmarinoes.smarthut.controller;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.when;

import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.SensorSaveRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.NotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.Sensor;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.SensorRepository;
import ch.usi.inf.sa4.sanmarinoes.smarthut.service.DeviceService;
import ch.usi.inf.sa4.sanmarinoes.smarthut.service.SensorService;
import java.math.BigDecimal;
import java.security.Principal;
import java.util.Optional;
import lombok.SneakyThrows;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.mock.web.MockHttpServletRequest;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

@DisplayName("The sensor controller")
@ExtendWith(MockitoExtension.class)
@WithMockUser(username = "user")
public class SensorControllerTests {
    @InjectMocks private SensorController sensorController;

    @Mock private DeviceService deviceService;

    @Mock private Principal mockPrincipal;

    @Mock private SensorRepository sensorRepository;

    @Mock private SensorService sensorService;

    @BeforeEach
    public void setup() {
        when(mockPrincipal.getName()).thenReturn("user");
    }

    private void checkSensorAgainstRequest(final Sensor toCheck, final SensorSaveRequest request) {
        assertThat(toCheck).isNotNull();
        assertThat(toCheck.getSensor()).isEqualTo(request.getSensor());
        assertThat(toCheck.getValue()).isEqualTo(request.getValue());
        assertThat(toCheck.getName()).isEqualTo(request.getName());
        assertThat(toCheck.getRoomId()).isEqualTo(request.getRoomId());
    }

    @DisplayName("when creating should return the same object")
    @Test
    @SneakyThrows(NotFoundException.class)
    public void testCreate() {
        doNothing().when(deviceService).throwIfRoomNotOwned(anyLong(), eq("user"));
        when(deviceService.saveAsOwner(any(Sensor.class), eq("user")))
                .thenAnswer(i -> i.getArguments()[0]);

        MockHttpServletRequest request = new MockHttpServletRequest();
        RequestContextHolder.setRequestAttributes(new ServletRequestAttributes(request));

        final SensorSaveRequest toSend =
                new SensorSaveRequest(
                        Sensor.SensorType.TEMPERATURE, BigDecimal.ZERO, 42L, "Test sensor");
        final Sensor created = sensorController.create(toSend, mockPrincipal);

        checkSensorAgainstRequest(created, toSend);
    }

    @DisplayName("when deleting an existing id should succeed")
    @Test
    @SneakyThrows(NotFoundException.class)
    public void testDelete() {
        doNothing().when(deviceService).deleteByIdAsOwner(eq(42L), eq("user"));

        MockHttpServletRequest request = new MockHttpServletRequest();
        RequestContextHolder.setRequestAttributes(new ServletRequestAttributes(request));

        Assertions.assertDoesNotThrow(() -> sensorController.deleteById(42L, mockPrincipal));
    }

    @Test
    public void testUpdateValue() throws NotFoundException {
        final Sensor s = new Sensor();
        when(sensorRepository.findByIdAndUsername(1L, "user")).thenReturn(Optional.of(s));
        when(sensorRepository.findByIdAndUsername(2L, "user")).thenReturn(Optional.empty());
        when(sensorService.updateValueFromSensor(s, BigDecimal.valueOf(30)))
                .thenAnswer(
                        i -> {
                            s.setValue(i.getArgument(1));
                            return s;
                        });
        assertThatThrownBy(() -> sensorController.updateValue(2L, BigDecimal.ZERO, mockPrincipal))
                .isInstanceOf(NotFoundException.class);
        assertThat(sensorController.updateValue(1L, BigDecimal.valueOf(30), mockPrincipal))
                .isSameAs(s);
    }

    @Test
    public void testUpdateSimulation() throws NotFoundException {
        final Sensor s = new Sensor();
        when(sensorRepository.findByIdAndUsername(1L, "user")).thenReturn(Optional.of(s));
        when(sensorRepository.findByIdAndUsername(2L, "user")).thenReturn(Optional.empty());
        when(sensorService.updateSimulationFromSensor(
                        s, BigDecimal.valueOf(30), BigDecimal.valueOf(10)))
                .thenAnswer(
                        i -> {
                            s.setError(i.getArgument(1));
                            s.setTypical(i.getArgument(2));
                            return s;
                        });
        assertThatThrownBy(() -> sensorController.updateValue(2L, BigDecimal.ZERO, mockPrincipal))
                .isInstanceOf(NotFoundException.class);
        assertThat(
                        sensorController.updateSimulation(
                                1L, BigDecimal.valueOf(30), BigDecimal.valueOf(10), mockPrincipal))
                .isSameAs(s);
    }
}
