package ch.usi.inf.sa4.sanmarinoes.smarthut.controller;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.when;
import static org.springframework.test.util.AssertionErrors.fail;

import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.SwitchableSaveRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.DuplicateStateException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.NotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.*;
import ch.usi.inf.sa4.sanmarinoes.smarthut.service.DeviceService;
import java.math.BigDecimal;
import java.security.Principal;
import java.util.Optional;
import lombok.SneakyThrows;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.security.test.context.support.WithMockUser;

@ExtendWith(MockitoExtension.class)
@WithMockUser(username = "user")
public class SmartPlugControllerTests {
    @InjectMocks private SmartPlugController smartPlugController;

    @Mock private UserRepository userRepository;

    @Mock private SmartPlugRepository smartPlugRepository;

    @Mock private Principal mockPrincipal;

    @Mock private DeviceService deviceService;

    @Mock private SceneRepository sceneRepository;

    @Mock private StateRepository<State> stateRepository;

    private final User u;

    public SmartPlugControllerTests() {
        u = new User();
        u.setName("user");
        u.setId(1L);
    }

    private void equalToRequest(Switchable created, SwitchableSaveRequest a) {
        assertThat(created.getName()).isEqualTo(a.getName());
        assertThat(created.getRoomId()).isEqualTo(30L);
        assertThat(created.isOn()).isEqualTo(a.isOn());
    }

    @Test
    public void testCreate() {
        when(mockPrincipal.getName()).thenReturn("user");
        when(deviceService.saveAsOwner(any(), eq("user"))).thenAnswer(i -> i.getArguments()[0]);

        SwitchableSaveRequest a = new SwitchableSaveRequest(true, 1L, 30L, "New SmartPlug");
        try {
            Switchable created = smartPlugController.create(a, mockPrincipal);
            assertThat(created.getId()).isEqualTo(0L);
            equalToRequest(created, a);
        } catch (NotFoundException e) {
            fail(e.getMessage());
        }
    }

    @Test
    public void testUpdate() throws NotFoundException {
        when(mockPrincipal.getName()).thenReturn("user");
        final SmartPlug old = new SmartPlug();
        old.setId(42L);

        when(deviceService.saveAsOwner(any(), eq("user"))).thenAnswer(i -> i.getArguments()[0]);
        when(smartPlugRepository.findByIdAndUsername(42L, "user")).thenReturn(Optional.of(old));
        when(smartPlugRepository.findByIdAndUsername(43L, "user")).thenReturn(Optional.empty());

        SwitchableSaveRequest a = new SwitchableSaveRequest(true, 42L, 30L, "New SmartPlug");

        SmartPlug created = smartPlugController.update(a, mockPrincipal);
        assertThat(created.getId()).isEqualTo(42L);
        equalToRequest(created, a);

        a.setId(43L);
        assertThatThrownBy(() -> smartPlugController.update(a, mockPrincipal))
                .isInstanceOf(NotFoundException.class);
    }

    @Test
    public void testDelete() throws NotFoundException {
        when(mockPrincipal.getName()).thenReturn("user");
        doNothing().when(deviceService).deleteByIdAsOwner(42L, "user");
        Assertions.assertDoesNotThrow(() -> smartPlugController.deleteById(42L, mockPrincipal));
    }

    @Test
    @SneakyThrows(NotFoundException.class)
    public void testResetMeter() {
        when(mockPrincipal.getName()).thenReturn("user");
        SmartPlug s = new SmartPlug();
        s.setId(24L);
        when(smartPlugRepository.findByIdAndUsername(24L, "user")).thenReturn(Optional.of(s));
        s.resetTotalConsumption();
        when(smartPlugRepository.save(s)).thenReturn(s);
        SmartPlug toCheck = smartPlugController.resetMeter(24L, mockPrincipal);
        assertThat(toCheck.getTotalConsumption()).isEqualTo(new BigDecimal(0));
    }

    @Test
    public void testSceneBinding() {
        when(mockPrincipal.getName()).thenReturn("user");
        SmartPlug smartPlug = new SmartPlug();
        when(smartPlugRepository.findByIdAndUsername(24L, "user"))
                .thenReturn(Optional.of(smartPlug));
        Scene scene = new Scene();
        scene.setId(1L);
        SwitchableState state = new SwitchableState();
        state.setSceneId(1L);
        State s = smartPlug.cloneState();
        when(sceneRepository.findById(1L)).thenReturn(Optional.of(scene));
        when(stateRepository.countByDeviceIdAndSceneId(24L, 1L)).thenReturn(0);
        Assertions.assertDoesNotThrow(
                () -> smartPlugController.sceneBinding(24L, 1L, mockPrincipal));
    }

    @Test
    public void testSceneBinding2() {
        when(mockPrincipal.getName()).thenReturn("user");
        SmartPlug smartPlug = new SmartPlug();
        when(smartPlugRepository.findByIdAndUsername(24L, "user"))
                .thenReturn(Optional.of(smartPlug));
        Scene scene = new Scene();
        scene.setId(1L);
        SwitchableState state = new SwitchableState();
        state.setSceneId(1L);
        State s = smartPlug.cloneState();
        when(sceneRepository.findById(1L)).thenReturn(Optional.of(scene));
        when(stateRepository.countByDeviceIdAndSceneId(24L, 1L)).thenReturn(2);
        assertThatThrownBy(() -> smartPlugController.sceneBinding(24L, 1L, mockPrincipal))
                .isInstanceOf(DuplicateStateException.class);
    }
}
