package ch.usi.inf.sa4.sanmarinoes.smarthut.controller;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

import ch.usi.inf.sa4.sanmarinoes.smarthut.config.EmailConfigurationService;
import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.InitPasswordResetRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.PasswordResetRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.dto.UserRegistrationRequest;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.DuplicateRegistrationException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.EmailTokenNotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.error.UserNotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.ConfirmationToken;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.ConfirmationTokenRepository;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.User;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.UserRepository;
import ch.usi.inf.sa4.sanmarinoes.smarthut.service.EmailSenderService;
import java.io.IOException;
import javax.servlet.http.HttpServletResponse;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.mail.SimpleMailMessage;
import org.springframework.mock.web.MockHttpServletResponse;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.test.context.support.WithMockUser;

@ExtendWith(MockitoExtension.class)
@WithMockUser(username = "user")
public class UserAccountControllerTests {

    @InjectMocks private UserAccountController userAccountController;

    @Mock private UserRepository userRepository;

    @Mock private EmailSenderService emailSenderService;

    @Mock private EmailConfigurationService emailConfigurationService;

    @Mock private ConfirmationTokenRepository confirmationTokenRepository;

    @Mock private BCryptPasswordEncoder passwordEncoder;

    @Test
    public void testRegisterUser() {
        final UserRegistrationRequest registrationRequest = new UserRegistrationRequest();
        registrationRequest.setEmail("info@theshell.ch");
        registrationRequest.setUsername("username");

        when(userRepository.findByEmailIgnoreCase("info@theshell.ch")).thenReturn(null);
        when(userRepository.findByEmailIgnoreCase("info@vimtok.com")).thenReturn(new User());
        when(userRepository.findByUsername("username")).thenReturn(new User());
        when(userRepository.findByUsername("simoneriva")).thenReturn(null);

        Assertions.assertThatThrownBy(() -> userAccountController.registerUser(registrationRequest))
                .isInstanceOf(DuplicateRegistrationException.class);

        registrationRequest.setUsername("simoneriva");
        registrationRequest.setEmail("info@vimtok.com");

        Assertions.assertThatThrownBy(() -> userAccountController.registerUser(registrationRequest))
                .isInstanceOf(DuplicateRegistrationException.class);
    }

    @Test
    public void testResetPassword() throws EmailTokenNotFoundException {
        final User u = new User();
        final ConfirmationToken c = new ConfirmationToken(u);
        c.setResetPassword(false);

        when(userRepository.save(u)).thenReturn(u);

        when(confirmationTokenRepository.findByConfirmToken("token")).thenReturn(c);
        when(confirmationTokenRepository.findByConfirmToken("token2")).thenReturn(null);
        doNothing().when(confirmationTokenRepository).delete(c);
        when(passwordEncoder.encode("password")).thenReturn("encoded");

        assertThatThrownBy(
                        () ->
                                userAccountController.resetPassword(
                                        new PasswordResetRequest("token2", "password")))
                .isInstanceOf(EmailTokenNotFoundException.class);

        assertThatThrownBy(
                        () ->
                                userAccountController.resetPassword(
                                        new PasswordResetRequest("token", "password")))
                .isInstanceOf(EmailTokenNotFoundException.class);

        c.setResetPassword(true);

        userAccountController.resetPassword(new PasswordResetRequest("token", "password"));

        assertThat(u.getPassword()).isEqualTo("encoded");
    }

    @Test
    public void testInitResetPassword() throws UserNotFoundException {
        when(emailConfigurationService.getResetPasswordSubject()).thenReturn("password reset");
        when(emailConfigurationService.getResetPassword()).thenReturn("reset-password");
        when(emailConfigurationService.getResetPasswordPath()).thenReturn("path");

        final User u = new User();
        u.setEmail("info@theshell.com");

        boolean[] done = new boolean[1];

        doAnswer(
                        i -> {
                            final SimpleMailMessage m = i.getArgument(0);
                            assertThat(m.getSubject()).isEqualTo("password reset");
                            assertThat(m.getText()).startsWith("reset-password path");
                            done[0] = true;
                            return null;
                        })
                .when(emailSenderService)
                .sendEmail(any());

        doNothing().when(confirmationTokenRepository).deleteByUserAndResetPassword(u, true);
        when(confirmationTokenRepository.save(any())).thenAnswer(i -> i.getArgument(0));

        when(userRepository.findByEmailIgnoreCase("info@theshell.ch")).thenReturn(u);
        when(userRepository.findByEmailIgnoreCase("info@vimtok.com")).thenReturn(null);

        assertThatThrownBy(
                        () ->
                                userAccountController.initResetPassword(
                                        new InitPasswordResetRequest("info@vimtok.com")))
                .isInstanceOf(UserNotFoundException.class);
        userAccountController.initResetPassword(new InitPasswordResetRequest("info@theshell.ch"));

        assertThat(done[0]).isTrue();
    }

    @Test
    public void testConfirmUserAccount() throws EmailTokenNotFoundException, IOException {
        final User u = new User();
        final ConfirmationToken c = new ConfirmationToken(u);
        c.setResetPassword(true);

        when(confirmationTokenRepository.findByConfirmToken("token")).thenReturn(null);
        when(confirmationTokenRepository.findByConfirmToken(c.getConfirmToken())).thenReturn(c);
        when(userRepository.save(u)).thenReturn(u);
        final HttpServletResponse r = new MockHttpServletResponse();

        assertThatThrownBy(() -> userAccountController.confirmUserAccount("token", r))
                .isInstanceOf(EmailTokenNotFoundException.class);
        assertThatThrownBy(() -> userAccountController.confirmUserAccount(c.getConfirmToken(), r))
                .isInstanceOf(EmailTokenNotFoundException.class);

        c.setResetPassword(false);

        when(emailConfigurationService.getRegistrationRedirect()).thenReturn("malusa.html");

        userAccountController.confirmUserAccount(c.getConfirmToken(), r);
        assertThat(u.isEnabled()).isTrue();
    }
}
